'use strict';

// This module manages Google aps

/* global appCore,IMatchModals,google,MarkerClusterer,IMatchTracks,CameraDatabase,MapUtility,moment,LatLon */

function GMapModule() {
    
    var self = this;

    self.gMap = null;
    self.markers = [];
    self.markerCluster = null;
    
    var files = [];
    var locations = [];

    var mapPopup = {
        // The one and only info window
        infoWnd: null,
        // If this is a cluster, the current index in that cluster
        index: 0,
        // The file coordinate
        gps: {},
        // The markers. If more than one element, this is a cluster
        markers : []
    }

    var markerTarget = null;
    var focusedMarker = null;

    self.popup = null;

    var selectTool = {
        // This contains the current object
        object : null,
        objectType: ''
    }

    var undo = {
        type : ''
    }

    var trackLogs = [];
    var trackBounds = null;
    var trackLogMarkerMinZoom = 14;

    var trackMarkers = [];

    /////////////////////////////////////////////////////////////////////////////
    // Dynamic load of Google Maps. We need to provide the user key
    // when the JavaScripts are loaded from the Google Server. 
    function loadGMaps(mapKey,callback) {
        if (!window.googleMapsLoaded) {
            var script = document.createElement('script');
            script.type = 'text\/JavaScript';
            document.body.appendChild(script);
            // Load the map and call callback when done
            script.src = 'https://maps.googleapis.com/maps/api/js?callback=' + callback + '&key=' + mapKey + '&language=' + appCore.IMatchConfig.locale + '&libraries=geometry';
        }
        else {
            if (callback) callback();
        }
    }    


    /////////////////////////////////////////////////////////////////////////////
    // This function is called by the Google API after the API has been loaded and initialized.
    // We now can load the pending map template.
    window.gmapsLoaded = function() {
        window.googleMapsLoaded = true;
        // Google API has been loaded. No load the template.
        var mt = appCore.getMapConfig($('#select-map-types').val());
        self.loadMap(false);
    }

    
    /////////////////////////////////////////////////////////////////////////////
    // Clear and reload the map
    this.loadMap = function(keepCenter)
    {
        // If the Google Maps API has not been loaded yet, we do it now.
        if (!window.googleMapsLoaded) {
            loadGMaps(appCore.getGMapsKey(),'gmapsLoaded');
            return;
        }
        else {
            closePopup();

            self.gMap = null;

            mapPopup.infoWnd = new google.maps.InfoWindow({});

            var center = new google.maps.LatLng(51.510121, -0.136357);

            if (appCore.appConfig.mapCenter !== undefined) {
                center = new google.maps.LatLng(appCore.appConfig.mapCenter.lat, appCore.appConfig.mapCenter.lng);
            }

            var options = {
                zoom: Math.min(21,Math.max(1,appCore.appConfig.zoomLevel)),
                center: center,
                scaleControl: true
               
                // rotateControl: true,
                // rotateControlOptions: {
                //     position: google.maps.ControlPosition.LEFT_CENTER
                // },
                // streetViewControl: true,
                // streetViewControlOptions: {
                //     position: google.maps.ControlPosition.RIGHT_TOP
                // },
                // panControl: true,
                // zoomControl: true,
                // zoomControlOptions: {
                //     position: google.maps.ControlPosition.RIGHT_CENTER
                // }
            };

            switch (appCore.getActiveMapConfig().type) {
                case 'google-satellite':
                    options.mapTypeId = 'satellite';
                    break;
                case 'google-street':
                    options.mapTypeId = 'roadmap';
                    break;
                case 'google-hybrid':
                    options.mapTypeId = 'hybrid';
                    break;
                case 'google-terrain':
                    options.mapTypeId = 'terrain';
                    break;
            }

            self.gMap = new google.maps.Map(document.getElementById('gmap'), options);

            // We watch for zoom change so we can update the user settings.
            self.gMap.addListener('zoom_changed', function () {
                saveConfig();
                window.setTimeout(function() {
                    createFOV();
                },500);
            });

            self.gMap.addListener('projection_changed', function () {
                // We need to wait for this in order to access self.gMap.getProjection()
                window.setTimeout(function() {
                    createFOV();
                },500);
            });
            

            // If we click somewhere on the map (outside a marker) close the info wnd.
            self.gMap.addListener('click', function (event) {
                closePopup();
                setTargetMarker(event.latLng.lng(),event.latLng.lat());
            });

            loadLocations();
            
            self.loadFiles(false);

            self.onTrackLoaded(false);
        }
    }


    /////////////////////////////////////////////////////////////////////////////
    // Returns true if the file with the given id is on the map.
    self.hasFile = function(id)
    {
        files.forEach(function(f) {
            if (f.id == id) return true;
        });
        return false;
    }

    var loadFilesDelayed = IMatch.debounce(function () {
        self.loadFiles(false);
    },2000);


    /////////////////////////////////////////////////////////////////////////////
    // Load the files in the file window
    self.loadFiles = function(keepCenter)
    {
        clearMap();
        files = [];

        var params = {
            fields: 'id,folderid,gps,name',
            rtagfocalLength: 'focallength',
            rtagfocallength35mm: 'focallength35mm',
            tagmake: 'make',
            tagmodel: 'model'
        }

        if (appCore.appConfig.showImages) {
            params.fields += ',imagewidth,imageheight,protected';
            params.imagesize = 'T';
        }

        var keepCenter = keepCenter || false;
        var center = null;
        var focusedFile = 0;

        // Get the focused file first. We use it to center the map.
        params.idlist = IMatch.idlist.fileWindowFocusedFile;
        IMWS.get('v1/files',params).then(function(response) {

            // If there is no focused file or the file has no GPS data we don't use it.
            if (response.files.length > 0) {
                if (response.files[0].gps !== undefined) {
                    focusedFile = response.files[0].id;
                    let fileCoords = MapUtility.getSafeCoordinates(focusedFile);
                    if (fileCoords != undefined) {
                        center = new google.maps.LatLng(fileCoords.lat, fileCoords.lon);
                    }
                }
            }

            // Now load the data for all files.
            params.idlist = IMatch.idlist.fileWindowSelection;
            IMWS.get('v1/files',params).then(function(response) {
                files = response.files;

                if (appCore.debugMode()) {
                    console.log('Gmaps loaded ' + response.files.length + ' files.');
                }

                // If we did not find a center yet, we use the first file with GPS data
                if (center == null) {
                    for (let i = 0; i < files.length; ++i) {
                        let fileCoords = MapUtility.getSafeCoordinates(files[i]);
                        if (fileCoords !== undefined) {
                            center = new google.maps.LatLng(fileCoords.lat,fileCoords.lon);
                            break;
                        }
                    }
                }

                createMarkers(files,focusedFile);

                if (!keepCenter && (center != null) && appCore.appConfig.followFocus) {
                    window.setTimeout(function () {
                        self.gMap.panTo(center);
                    }, 50);
                }

                appCore.checkCameraDB();
            });
        });
    }

    /////////////////////////////////////////////////////////////////////////////
    // Clear the Undo
    function clearUndo()
    {
        undo = {
            type : ''
        }
        appCore.setUndo(false);
    }


    /////////////////////////////////////////////////////////////////////////////
    // Clear the map
    function clearMap()
    {
        clearUndo();
        clearMarkers(true);
        closePopup();
    }

    
    /////////////////////////////////////////////////////////////////////////////
    // Clear all markers
    function clearMarkers(keepTargetMarker) {
        // Clear existing markers
        setFocusedMarker(null);
        var n = self.markers.length;
        var i;
        for (i = 0; i < n; ++i) {
            var m = self.markers[i];
            if (m.fov) {
                if (m.fov.poly) {
                    m.fov.poly.setMap(null);
                }
            }
            if (m.dest) {
                if (m.dest.marker) {
                    m.dest.marker.setMap(null);
                }
                if (m.dest.link) {
                    m.dest.link.setMap(null);
                }
            }
            m.setMap(null);            
        }
        self.markers = [];

        if (self.markerCluster) {
            self.markerCluster.clearMarkers();
        }

        keepTargetMarker = keepTargetMarker || false;

        if (markerTarget && !keepTargetMarker) {
            markerTarget.setMap(null);
            markerTarget = null;
            appCore.onTargetMarkerChanged(false);
        }
    }
    

    /////////////////////////////////////////////////////////////////////////////
    // Clear all locations
    function clearLocations()
    {
        locations.forEach(function(l) {
            l.marker.setMap(null);
            l.setMap(null);
        });
        locations = [];
    }

    function rotatePolygon(polygon,angle) {
        var map = self.gMap;
        var prj = map.getProjection();
        if (prj) {
            var origin = prj.fromLatLngToPoint(polygon.getPath().getAt(0)); //rotate around first point
        
            var coords = polygon.getPath().getArray().map(function(latLng){
                var point = prj.fromLatLngToPoint(latLng);
                var rotatedLatLng =  prj.fromPointToLatLng(MapUtility.rotatePoint(point,origin,angle));
                return {lat: rotatedLatLng.lat(), lng: rotatedLatLng.lng()};
            });

            polygon.setPath(coords);
        }
    }    
            

    /////////////////////////////////////////////////////////////////////////////
    // Create markers for all loaded files
    function createMarkers(files,focus)
    {
        var showdata = appCore.appConfig.showFileData;
        var showimg = appCore.appConfig.showImages;

        files.forEach(function (f) {
            let fileCoords = MapUtility.getSafeCoordinates(f);
            if (fileCoords !== undefined) {

                let options = {
                    map: self.gMap,
                    coordType: fileCoords.type,
                    file: f,
                    title: f.name,
                    position: new google.maps.LatLng(fileCoords.lat, fileCoords.lon),
                    // Only when not locked
                    draggable: !appCore.appConfig.lockMove
                }
               
                if (showdata) {
                    options.label = {
                        text: f.name
                    }

                    // If we display a hybrid or street map we 
                    // change the text color to white to make it better visible
                    switch (appCore.getActiveMapConfig().type) {
                        case 'google-satellite':
                            options.label.color =  '#ffffff';
                            break;
                        case 'google-street':
                            break;
                        case 'google-hybrid':
                            options.label.color =  '#ffffff';
                            break;
                        case 'google-terrain':
                            break;
                    }                    
                }

                if (showimg) {
                    var prot = f.protected === true;
                    var w = f.imageWidth;
                    var h = f.imageHeight;
                    if (prot) {
                        w = h = 32;
                    }
                    var s = Math.min(1.0, 80.0 / Math.max(1, w));

                    options.icon = {
                        url: prot ? appCore.shieldIconUrl() : appCore.fileUrl(f.id),
                        size: new google.maps.Size(w*s, h*s), //! Else the hit testing is using the original (300 pixel size!)
                        scaledSize: new google.maps.Size(w * s, h * s),
                        origin: new google.maps.Point(0,0),
                        anchor: new google.maps.Point((w*s)/2,(h*s)),
                        labelOrigin: new google.maps.Point((w*s)/2,h*s+10),
                    }
                }
                else {
                    // We use a custom icon. This makes the markers look the 
                    // same as in OL and we can position the label below the icon!
                    options.icon = {
                        // url: f.id == focus ? 'img/marker2.svg' : 'img/marker.svg',
                        // labelOrigin: new google.maps.Point(0,40),
                        // scaledSize: new google.maps.Size(20,30),

                        // Cannot use SVG files. Google Maps changes the scale (?)
                        // when the draggable option changes... :-/
                        url: 'img/marker.png',
                        labelOrigin: new google.maps.Point(0,40)
                    }
                }

                var marker = new google.maps.Marker(options);

                if (f.id == focus) {
                    setFocusedMarker(marker);
                }

                // Does the file have a dest GPS record?
                // If so, add the destination marker.
                if (f.gps.destLat && f.gps.destLon) {
                    createDestinationMarker(marker);
                }

                self.markers.push(marker);

                marker.addListener('click', function (event) {
                    setFocusedMarker(marker,true);
                });

                marker.addListener('rightclick', function (event) {
                    closePopup();
                    setFocusedMarker(marker,true);
                    createInfoWndContent([marker]);
                });

                // Add support for dragging
                if (!appCore.appConfig.lockMove) {
                    marker.addListener('dragend', function (event) {

                        undo.type = 'file';
                        // Deep copy
                        undo.file = $.extend(true,{},this.file);
                        appCore.setUndo(true);

                        let params = {
                            target: marker.coordType,
                            id: this.file.id
                        }
                        if (marker.coordType == 'created') {
                            params.lng = event.latLng.lng(),
                            params.lat = event.latLng.lat()
                        }
                        else {
                            params.destlng = event.latLng.lng(),
                            params.destlat = event.latLng.lat()
                        }
                        IMatch.setGPSCoordinates(params).then(response => {
                            // Fetch the data of the updated file. 
                            // IMatch will fill in image direction and other values.
                            IMWS.get('v1/files',{
                                id: marker.file.id,
                                fields: 'gps'
                            }).then(response => {
                                marker.file.gps = response.files[0].gps;
                                createDestinationMarker(marker);
                                createFOV();
                            });
                        },error => {
                            appCore.showError('setGps',error);
                        });

                    });
                }
            }
        },this);

        var options = {
            imagePath: 'gm/gmaps-clusterer/images/m',
            gridSize: 5,  // Match to feature clusterer in ol
            minimumClusterSize: 2,
            zoomOnClick : false
        };

        if (self.markerCluster == null) {
            self.markerCluster = new MarkerClusterer(self.gMap, self.markers, options);

            self.markerCluster.addListener('clusterclick', function (cluster, event) {
                closePopup();
                createInfoWndContent(cluster.getMarkers(),cluster.getCenter());
                // Important!
                event.cancelBubble = true;
                if (event.stopPropagation) {
                    event.stopPropagation();
                }        
                return false;
            });
        }
        else {
            self.markerCluster.clearMarkers();
            self.markerCluster.addMarkers(self.markers);
        }

        window.setTimeout(function() {
            createFOV();
        },500);
    }    

    
    /////////////////////////////////////////////////////////////////////////////
    // Creates a destination marker for the specified parent marker.
    function createDestinationMarker(parentMarker)
    {
        if (parentMarker.dest != undefined) {
            if (parentMarker.dest.marker) {
                parentMarker.dest.marker.setMap(null);
            }
            if (parentMarker.dest.link) {
                parentMarker.dest.link.setMap(null);
            }
        }
        
        if (parentMarker.file.gps.lat && parentMarker.file.gps.lon && parentMarker.file.gps.destLat && parentMarker.file.gps.destLon) {
            let options = {
                map: self.gMap,
                file: parentMarker.file,
                position: new google.maps.LatLng(parentMarker.file.gps.destLat, parentMarker.file.gps.destLon),
                // Only when not locked
                draggable: !appCore.appConfig.lockMove,
                parentMarker: parentMarker,
                icon : {
                    url: 'img/destination.png',
                    anchor: new google.maps.Point(15,15)
                }
            }

            parentMarker.dest = {
                marker : new google.maps.Marker(options),
                link: new google.maps.Polyline({
                    map: self.gMap,
                    path: [{lat: parentMarker.file.gps.lat, lng: parentMarker.file.gps.lon},{lat: parentMarker.file.gps.destLat, lng: parentMarker.file.gps.destLon}],
                    geodesic: true,
                    strokeColor: '#FF0000',
                    strokeOpacity: 1.0,
                    strokeWeight: 4
                })
            }

            // Add support for dragging
            if (!appCore.appConfig.lockMove) {
                parentMarker.dest.marker.addListener('dragend', function (event) {

                    undo.type = 'dest';
                    // Deep copy
                    undo.file = $.extend(true,{},this.file);
                    appCore.setUndo(true);

                    let ll = event.latLng;

                    // if (this.file.gps.lat && this.file.gps.lon && this.file.gps.destLat && this.file.gps.destLon) {
                    //     let dir = google.maps.geometry.spherical.computeHeading(new google.maps.LatLng(this.file.gps.lat,this.file.gps.lon), new google.maps.LatLng(ll.lat(),ll.lng()));
                    //     MapUtility.setImgDirection(this.file.id,dir,'T');
                    // }

                    IMatch.setGPSCoordinates({
                        target: 'dest',
                        destlng: event.latLng.lng(),
                        destlat: event.latLng.lat(),
                        id: this.file.id
                    }).then(response => {
                        // Fetch the data of the updated file. 
                        // IMatch will fill in image direction and other values.
                        IMWS.get('v1/files',{
                            id: parentMarker.file.id,
                            fields: 'gps'
                        }).then(response => {
                            parentMarker.file.gps = response.files[0].gps;
                            createDestinationMarker(parentMarker);
                            createFOV();
                        });
                    },error => {
                        appCore.showError('setGps',error);
                    });
                });

                parentMarker.dest.marker.addListener('rightclick', function (event) {
                    closePopup();
                    createDestinationMarkerPopup(event.latLng,this);
                });

                // Make the parent active when we click the dest marker
                parentMarker.dest.marker.addListener('click', function (event) {
                    setFocusedMarker(this.parentMarker,true);
                });
                
                
            }
        }
    }


    /////////////////////////////////////////////////////////////////////////////
    //
    function setFocusedMarker(marker,updateFileWindow)
    {
        if (focusedMarker == marker) return;
        updateFileWindow = updateFileWindow || false;
        
        if (focusedMarker != null) {
            if (!appCore.appConfig.showImages) {
                let i = focusedMarker.getIcon();
                i.url = 'img/marker.png';
                focusedMarker.setIcon(i);
            }
        }
        focusedMarker = marker;
        
        if (focusedMarker != null) {
            if (!appCore.appConfig.showImages) {
                let i = focusedMarker.getIcon();
                i.url = 'img/marker-focus.png';
                focusedMarker.setIcon(i);
            }
            if (updateFileWindow) {
                window.setTimeout(function() {
                    IMatch.focusFile({
                        id: focusedMarker.file.id
                    });
                },100);
            }
            createFOV();
        }
    }


    /////////////////////////////////////////////////////////////////////////////
    // Create/Update the FOV for all files or the focused file, depending on 
    // the current settings.
    function createFOV()
    {
        if (!focusedMarker) return;
        
        var n = self.markers.length;
        for (let i = 0; i < n; ++i) {
            let m = self.markers[i];
            
            // Clear existing FOV
            if (m.fov) {
                if (m.fov.poly) {
                    m.fov.poly.setMap(null);
                }
                m.fov = null;
            }

            // If we show all FOVs or this is the focused marker
            if ((appCore.appConfig.fieldOfView == 'all') || ((appCore.appConfig.fieldOfView == 'focused') && (m.file.id == focusedMarker.file.id))) {

                // We can generate a FOV when we have gos.imgDirection, with a fallback to gps.track.
                let dir = m.file.gps.imgDirection;
                if (dir == undefined) {
                    dir = m.file.gps.track;
                }
                if (dir == undefined) {
                    // If we have a destination point, we can use this as a 2nd fallback
                    if (m.file.gps.lat && m.file.gps.lon && m.file.gps.destLat && m.file.gps.destLon) {
                        dir = google.maps.geometry.spherical.computeHeading(new google.maps.LatLng(m.file.gps.lat,m.file.gps.lon), new google.maps.LatLng(m.file.gps.destLat,m.file.gps.destLon));
                    }
                }

                if (dir != undefined) {
                    var pmarker = LatLon(m.file.gps.lat, m.file.gps.lon);

                    let cropFactor = 1.0;

                    // We need a make/model for a proper crop factor
                    if (m.file.make && m.file.model) {
                        let camInfo = appCore.camDB.getCameraInfo(m.file.make,m.file.model);
                        if (camInfo.result == 'ok') {
                            cropFactor = camInfo.cropFactor;
                        }
                        else if (camInfo.result != 'unknown') {
                            // This model has not set been tested.
                            appCore.camDB.lookupCameraInfo(m.file.make,m.file.model).then(response => {
                                if (response.result == 'ok') {
                                    // The model has been found.
                                    createFOV();
                                }
                            });
                        }
                    }

                    // We need one of the focal length to create the FOV
                    if (m.file.focalLength || m.file.focallength35mm) {
                        let flen35 = undefined;
                        if (parseFloat(m.file.focallength35mm) > 0) {
                            flen35 = parseFloat(m.file.focallength35mm);
                        }
                        let fov = MapUtility.getFOV(cropFactor,m.file.focalLength,flen35);

                        var bounds = self.gMap.getBounds();
                        let fovSize = 1000;
                        if (bounds) {
                            var p = LatLon(bounds.getNorthEast().lat(), bounds.getNorthEast().lng());
                            fovSize = Math.min(50000,pmarker.distanceTo(p)*2);
                        }

                        var pdest1 = pmarker.destinationPoint(fovSize, -fov/2);
                        var pdest2 = pmarker.destinationPoint(fovSize, +fov/2);

                        var coords = [
                            {lat: pmarker.lat, lng: pmarker.lon},
                            {lat: pdest1.lat, lng: pdest1.lon},
                            {lat: pdest2.lat, lng: pdest2.lon},
                        ];
                        // Close Poly
                        coords.push(coords[0]);
                
                        // Construct the polygon.
                        var fovPoly = new google.maps.Polygon({
                            paths: coords,
                            strokeColor: '#0000ff',
                            strokeOpacity: 0.4,
                            strokeWeight: 2,
                            fillColor: '#0000ff',
                            fillOpacity: 0.25
                        });

                        rotatePolygon(fovPoly,dir);

                        fovPoly.setMap(self.gMap);

                        m.fov = {
                            poly: fovPoly
                        }
                    }
                }
            }
        }

    }


    /////////////////////////////////////////////////////////////////////////////
    // 
    function closePopup() {
        if (mapPopup.infoWnd != null) {
            mapPopup.infoWnd.close();
            mapPopup.index = 0;
        }
    }

    /////////////////////////////////////////////////////////////////////////////
    // Move to the next/previous file in the popup
    function filePopupNext(dir) {
        if (mapPopup.markers == null) return;

        var cur = mapPopup.index;

        if (dir == 1) {
            if (mapPopup.index + 1 < mapPopup.markers.length) {
                mapPopup.index++;
            }
            else {
                mapPopup.index = 0;
            }
        }
        else {
            if (mapPopup.index > 0) {
                mapPopup.index--;
            }
            else if (mapPopup.markers.length > 0) {
                mapPopup.index = mapPopup.markers.length - 1;
            }
        }

        createInfoWndContent(mapPopup.markers);
    }

    /////////////////////////////////////////////////////////////////////////////
    // Markers is an array of 1 to n markers
    function createInfoWndContent(markers,centerPoint) {
        mapPopup.markers = markers;
        mapPopup.index = Math.min(mapPopup.index, markers.length - 1);

        // Fetch the data for the current file
        var id = mapPopup.markers[mapPopup.index].file.id;

        // Load data for the popup file from IMWS
        IMWS.get('v1/files',
            {
                id: id,
                fields: 'name,ext,imagewidth,imageheight,rating,label,labelcolor,protected,gps',
                imagesize: 'T',
                tagdesc: "description",
                tagheadline: "headline",
                tagtitle: "title",
                tagcountrycode: "countrycode",
                tagcountry: "country",
                tagstate: "state",
                tagcity: "city",
                taglocation: "location"
                // We cannto use gpsposition because that's a Composite tag
                // and it is not filled until the file has been written back.
                // If we change coordinates in the Map panel, this will not be the case.
                // tagalt: "gpsaltitude",
                // tagposition: "gpsposition",
            }).then(function (data) {
                if (data.files.length > 0) {
                    var file = data.files[0];

                    // We keep the precise location of this file
                    mapPopup.gps = file.gps;
                    
                    // NO: We show the decimal coordinates. They are always filled, even if the file has not yet been written back.
                    // var pos = file.lat.replace('deg', '°') + ', ';
                    // pos += file.lng.replace('deg', '°');
                    
                    // We show the decimal coordinates. They are always filled, even if the file has not yet been written back.
                    var pos;
                    if (file.gps.lat && file.gps.lon) {
                        pos = file.gps.lat + ', ' + file.gps.lon;
                    }
                    else if (file.gps.destLat && file.gps.destLon) {
                        pos = file.gps.destLat + ', ' + file.gps.destLon;
                    }
                    else {
                        pos = '';
                    }

                    var content = [
                        '<div id="popup" class="ptc-map-popup-gmaps container-fluid">',
                        '<img class="col-xs-6 thumb" src="' + (file.protected ? appCore.shieldIconUrl(true) : appCore.fileUrl(id)) + '" />',
                        '  <div class="row">',
                        '    <div class="fname">' + file.name + '</div>',
                        '    <div class="coord">' + pos + '</div>',
                        '    <div class="data">' + IMatch.safeTagValue(file.desc) + '</div>',
                        '  </div>',

                        // Rating & Label
                        '<div class="colctrl">',
                        '    <span class="rating">{rating}</span>',
                        '    <span ng-show="file.label.length" class="label badge" style="background-color: ' + IMatch.safeTagValue(file.labelColor) + '" title="' + IMatch.safeTagValue(file.label) + '">&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</span>',
                        '</div>'
                    ].join('\n');

                    // Set the rating

                    if (file.rating == -1) {
                        // Show the reject icon
                        content = content.replace('{rating}','<i class="ratingitem fa fa-ban fa-lg"></i>');
                    }
                    else if (file.rating > 0) {
                        // Show the stars
                        var s = '';
                        for (var i = 0; i < file.rating; ++i) {
                            s += '<i class="ratingitem fa fa-star fa-lg"></i>';
                        }
                        content = content.replace('{rating}',s);
                    }
                    else {
                        content = content.replace('{rating}','');
                    }

                    if ((file.country != undefined) || (file.city != undefined) || (file.location != undefined)) {
                        var ftr = [
                            '  <div class="row">',
                            '    <div class="footer">',
                            '      <span>' + IMatch.safeTagValue(file.country) + '</span><br/>',
                            '      <span>' + IMatch.safeTagValue(file.city) + '</span><br/>',
                            '      <span>' + IMatch.safeTagValue(file.location) + '</span><br/>',
                            '    </div>',
                            '  </div>'
                        ].join('\n');

                        content = content.concat(ftr);
                    }

                    // If this is a cluster, we add controls to navigate between the different files
                    if (mapPopup.markers.length > 1) {
                        var nav = [
                            '  <div class="row">',
                            '    <div class="navigator">',
                            '      <a id="popup-prev" href="#"><i class="fa fa-chevron-circle-left fa-2x ptc-m-r-10"></i></a>',
                            '      <a id="popup-next" href="#"><i class="fa fa-chevron-circle-right fa-2x ptc-m-r-10"></i></a>',
                            // n of m files (cluster)
                            '      <span class=""> ' + (mapPopup.index + 1) + ' / '+ mapPopup.markers.length + '</span>',
                            '    </div>',
                            '  </div>'
                        ].join('\n');
                        content = content.concat(nav);
                    }

                    // The toolbar at the bottom

                    var btnToolbar = [
                        '<div class="row">',
                        '  <div class="btn-toolbar pull-right">',
                        '    <button id="btn-target-marker" class="btn btn-default btn-sm" data-imt-title="MAP_POPUP_BTN_TARGET"><img src="img/target2.svg" class="ma-fa-icon"></button>',
                        '    <button id="btn-create-loc" class="btn btn-default btn-sm" data-imt-title="MAP_POPUP_BTN_LOC"><img src="img/location.svg" class="ma-fa-icon"></button>',
                        '    <button id="btn-create-dest" class="btn btn-default btn-sm" data-imt-title="MAP_POPUP_BTN_DIRECTION"><img src="img/destination.svg" class="ma-fa-icon"></button>',
                        '    <button id="btn-goto" class="btn btn-default btn-sm" data-imt-title="MAP_POPUP_BTN_GOTO"><i class="fa fa-th-large"></i></button>',
                        '    <button id="btn-find-nearby" class="btn btn-default btn-sm" data-imt-title="MAP_POPUP_BTN_FINDNEARBY"><i class="fa fa-search"></i></button>',
                        '    <button id="btn-reverse-geocode" class="btn btn-default btn-sm" data-imt-title="MAP_POPUP_BTN_GEOCODE"><i class="fa fa-globe"></i></button>',
                        '    <button id="btn-popup-close" class="btn btn-default btn-sm" style="margin-left:20px;" data-imt-title="MAP_POPUP_BTN_CLOSE"><i class="fa fa-times"></i></button>',
                        '  </div>',
                        '</div>',
                    ].join('\n');

                    content = content.concat(btnToolbar);
                    content = content.concat('\n</div>');

                    mapPopup.marker = mapPopup.markers[mapPopup.index];
                    mapPopup.infoWnd.setContent(content);

                    if (centerPoint !== undefined) {
                        mapPopup.infoWnd.open(self.gMap);
                        mapPopup.infoWnd.setPosition(centerPoint);
                    }
                    else {
                        mapPopup.infoWnd.open(self.gMap, mapPopup.marker);
                    }

                    // Translate the contents in the popup after it has been added to the DOM.
                    IMatchTranslate.translateElement('#popup');

                    // Button handlers

                    $('#btn-popup-close').click(function(e) {
                        closePopup();
                        e.preventDefault();
                        return false;
                    });

                    $('#btn-target-marker').click(function(e) {
                        // We use the current marker shown in the popup.
                        // If this is a cluster, index may be > 0
                        var pos = mapPopup.markers[mapPopup.index].getPosition();
                        setTargetMarker(pos.lng(),pos.lat());
                        e.preventDefault();
                        return false;
                    });
                    
                    $('#btn-create-loc').click(function(e) {
                        var m = mapPopup.markers[mapPopup.index];
                        var pos = m.getPosition();
                        IMatch.createGeoLocation({
                            lat: pos.lat(),
                            lng: pos.lng(),
                            templatefile: m.file.id
                        });
                        e.preventDefault();
                        return false;
                    });

                    $('#btn-create-dest').click(function(e) {
                        var m = mapPopup.markers[mapPopup.index];
                        createDirection(m);
                        e.preventDefault();
                        return false;
                    });

                    $('#btn-goto').click(function(e) {
                        var m = mapPopup.markers[mapPopup.index];
                        IMatch.gotoFolder(null,m.file.id);
                        e.preventDefault();
                        return false;
                    });

                    $('#btn-find-nearby').click(function(e) {
                        var pos = mapPopup.markers[mapPopup.index].getPosition();
                        IMatch.findNearbyFiles({
                            lat: pos.lat(),
                            lng: pos.lng(),
                            alt: 0                
                        }).then(function(response) {
                            if (response.result == 'ok') {
                                appCore.openFindNearbyResultWindow(IMatchTranslate.getStr('MAP_FIND_RESULTWINDOW'),response.id);
                            }
                        });
                        e.preventDefault();
                        return false;
                    })

                    $('#btn-reverse-geocode').click(function(e) {
                        var m = mapPopup.markers[mapPopup.index];
                        IMatch.reverseGeoCode({
                            id: m.file.id
                        }).then(function(response) {
                        });
                        e.preventDefault();
                        return false;
                    })
                    
                    $('#popup-prev').click(function(e) {
                        filePopupNext(0);
                        e.preventDefault();
                        return false;
                    })

                    $('#popup-next').click(function(e) {
                        filePopupNext(1);
                        e.preventDefault();
                        return false;
                    })                    

                    return content;
                }
            });

        return '';
    }

    /////////////////////////////////////////////////////////////////////////////
    // The popup for the destination marker
    function createDestinationMarkerPopup(latLng,marker)
    {
        closePopup();

        var content = [
            `<div id="target-popup-gmaps" class="ptc-map-popup-gmaps container-fluid">
              <h4 imt-translate="MAP_POPUP_DEST_MARKER"></h4>
              <div class="row">
                <div class="col-sm-12 ptc-popup-row"><i class="fa fa-check-square fa-lg imatch-margin-r-10"></i><a href="#" id="btn-dest-apply" data-imt-translate="MAP_POPUP_BTN_DEST_APPY"></a></div>
              </div>
            </div>`
        ].join('\n');

        mapPopup.infoWnd.setContent(content);
        mapPopup.infoWnd.setOptions({});
        mapPopup.infoWnd.open(self.gMap);
        mapPopup.infoWnd.setPosition(latLng);
        mapPopup.destMarker = marker;

        // Translate the contents in the popup after it has been added to the DOM.
        IMatchTranslate.translateElement('#target-popup-gmaps');

        // Button handlers

        $('#btn-popup-close').click(function(e) {
            closePopup();
            e.preventDefault();
            return false;
        })

        $('#btn-dest-apply').click(function(e) {
            closePopup();

            IMatch.setGPSCoordinates({
                idlist: IMatch.idlist.fileWindowSelection,
                target: 'dest',
                destlat: mapPopup.destMarker.getPosition().lat(),
                destlng: mapPopup.destMarker.getPosition().lng()
            }).then(response => {
                self.loadFiles(true);
            },error => {
                appCore.showError('setGps',error);
            });

            return false;
        })

        return true;
    }

    
    /////////////////////////////////////////////////////////////////////////////
    // Load all locations from the database
    function loadLocations()
    {
        clearLocations();
        
        IMWS.get('v1/geolocation').then(function(response) {

            response.locations.forEach(function(l) {

                var shape = new google.maps.Circle({
                    strokeColor: '#ff0000',
                    strokeOpacity: 0.75,
                    strokeWeight: 4,
                    fillColor: '#ffff00',
                    fillOpacity: 0.5,
                    map: self.gMap,
                    center: new google.maps.LatLng(l.lat, l.lng),
                    radius: l.mapRadius,
                    editable: appCore.appConfig.showLocations && !appCore.appConfig.lockMove,
                    draggable: appCore.appConfig.showLocations && !appCore.appConfig.lockMove,
                    visible: appCore.appConfig.showLocations,
                    imatchLocation: l
                });

                // To display the name of the location, we add an invisible marker.
                // Only the label assigned to the marker is shown in the center of
                // the location circle.
                shape.marker = new google.maps.Marker({
                    map: self.gMap,
                    icon: {
                        url: 'img/marker.svg',
                        scaledSize: new google.maps.Size(0,0),
                        labelOrigin: new google.maps.Point(0,0),
                    },
                    label: l.name,
                    title: l.name,
                    position: new google.maps.LatLng(l.lat, l.lng),
                });


                shape.addListener('rightclick', function (event) {
                    closePopup();
                    createLocationPopup(event.latLng,this);
                });

                if (!appCore.appConfig.lockMove) {
                    shape.addListener('dragend', function (event) {
                        updateLocation(this);
                    });
                    shape.addListener('radius_changed', function (event) {
                        updateLocation(this);
                    });
                    shape.addListener('center_changed', function (event) {
                        updateLocation(this);
                    });


                    function updateLocation(location)
                    {
                        undo.type = 'location';
                        // Deep copy
                        undo.location = $.extend(true,{},location.imatchLocation);
                        appCore.setUndo(true);

                        var center = location.getCenter();
                        var radius = location.getRadius();

                        // Move the marker to the same position!
                        location.marker.setPosition(center);

                        IMWS.post('v1/geolocation/update',{
                            guid: location.imatchLocation.guid,
                            lng: center.lng(),
                            lat: center.lat(),
                            radius: radius,
                        }).then(function(response) {
                        });
                    }
                }

                locations.push(shape);
            });
        });
    }
    

    /////////////////////////////////////////////////////////////////////////////
    // Undo the last operation
    this.onUndo = function() {
        if (undo.type == 'location') {
            IMWS.post('v1/geolocation/update',{
                guid: undo.location.guid,
                lng: undo.location.lng,
                lat: undo.location.lat,
                radius: undo.location.mapRadius,
            }).then(function(response) {
                if (response.result != 'ok') {
                    loadLocations();
                }
            });
        }
        else if (undo.type == 'file') {
            IMatch.setGPSCoordinates({
                target: 'created',
                lng: undo.file.gps.lon,
                lat: undo.file.gps.lat,
                id: undo.file.id
            }).then(function(response) {
                self.loadFiles(true);
            },error => {
                appCore.showError('setGps',error);
            });
        }
        else if (undo.type == 'dest') {
            IMatch.setGPSCoordinates({
                target: 'dest',
                destlng: undo.file.gps.destLon,
                destlat: undo.file.gps.destLat,
                id: undo.file.id
            }).then(function(response) {
                self.loadFiles(true);
            },error => {
                appCore.showError('setGps',error);
            });
        }
        clearUndo();
    }


    /////////////////////////////////////////////////////////////////////////////
    // Show all files
    this.onZoomToExtent = function() {
        if (self.markers.length == 0) return;
        
        var bounds = new google.maps.LatLngBounds();
        var n = self.markers.length;
        var i;
        for (i = 0; i < n; ++i) {
            bounds.extend(self.markers[i].getPosition());
            // Also consider the destination marker.
            if (self.markers[i].dest !== undefined) {
                bounds.extend(self.markers[i].dest.marker.getPosition());
            }
        }
        self.gMap.fitBounds(bounds);
        self.gMap.panToBounds(bounds);
    }

    /////////////////////////////////////////////////////////////////////////////
    // Let the user draw a selection
    this.onSelectTool = function()
    {
        closePopup();
        removeSelectTool();

        appCore.switchToolbar('select');

        // Remember
        selectTool.objectType = appCore.appConfig.selectTool;

        var centerTo = null;

        // 1° longitude at equator level is about 111 m.
        // We use this small calculation to create a reference 'meter'
        var meter = (self.gMap.getBounds().getNorthEast().lng() - self.gMap.getBounds().getSouthWest().lng()) * 111;

        if (appCore.appConfig.selectTool == 'circle') {

            var center = self.gMap.getCenter();
            var radius =  meter * 100;
            if (selectTool.circle !== undefined) {
                // We use the stored radius, but center to the current map center
                radius = selectTool.circle.radius;
            }

            selectTool.object = new google.maps.Circle({
                strokeColor: '#d82626',
                strokeOpacity: 0.8,
                strokeWeight: 4,
                fillColor: '#32a0cc',
                fillOpacity: 0.35,
                map: self.gMap,
                center: center,
                radius: radius,
                draggable: true,
                editable: true
            });

            centerTo = selectTool.object.getCenter();
        }
        else if (appCore.appConfig.selectTool == 'rectangle') {

            // ~ 1/10 of the visible map width
            var d = meter / 111 / 10;

            var center = self.gMap.getCenter();
            var bounds = {
                north: center.lat(),
                south: center.lat()+d,
                east: center.lng()+d,
                west: center.lng(),
            }
            if (selectTool.rect !== undefined) {

                // Use the stored dimensions and shift to the new center

                var dlat = center.lat() - selectTool.mapCenter.lat();
                var dlng = center.lng() - selectTool.mapCenter.lng();
                
                bounds.north = selectTool.rect.bounds.getNorthEast().lat() + dlat;
                bounds.south = selectTool.rect.bounds.getSouthWest().lat() + dlat;
                bounds.east = selectTool.rect.bounds.getNorthEast().lng() + dlng;
                bounds.west = selectTool.rect.bounds.getSouthWest().lng() + dlng;
            }

            selectTool.object = new google.maps.Rectangle({
                strokeColor: '#d82626',
                strokeOpacity: 0.8,
                strokeWeight: 4,
                fillColor: '#32a0cc',
                fillOpacity: 0.35,
                map: self.gMap,
                bounds: bounds,
                draggable: true,
                editable: true
            });

            centerTo = selectTool.object.getBounds().getCenter();
        }
        else if (appCore.appConfig.selectTool == 'polygon') {
            
            var center = self.gMap.getCenter();
           
            // ~ 1/20 of the visible map width
            var d = meter / 111 / 20;
           
            var path = [
                {lat: center.lat(), lng: center.lng() - d},
                {lat: center.lat()-d, lng: center.lng()+d},
                {lat: center.lat(), lng: center.lng()+d},
            ]

            if (selectTool.poly !== undefined) {

                var dlat = center.lat() - selectTool.mapCenter.lat();
                var dlng = center.lng() - selectTool.mapCenter.lng();

                // Shift the polygon to the current map center.
                path = [];
                selectTool.poly.path.forEach(function(n) {
                    path.push({
                        lat: n.lat() + dlat, 
                        lng: n.lng() + dlng
                    });
                });

            }

            selectTool.object = new google.maps.Polygon({
                strokeColor: '#d82626',
                strokeOpacity: 0.8,
                strokeWeight: 4,
                fillColor: '#32a0cc',
                fillOpacity: 0.35,
                map: self.gMap,
                path: path,
                draggable: true,
                editable: true
            });

            // Find the center of the polygon
            var bounds = new google.maps.LatLngBounds();
            selectTool.object.getPath().forEach(function(n, number) {
                bounds.extend(n);
            });

            // We handle the right-click for polygons to allow the user to remove nodes.
            selectTool.object.addListener('rightclick', function(event) {
                if (event.vertex != undefined) {
                    var path = this.getPath();
                    if (path.length > 3) {
                        path.removeAt(event.vertex);                    
                    }
                }
            });            
            
            centerTo = bounds.getCenter();
        }

        // Keep track of the map center relative to the position of the object.
        // We need this to shift the object to the new center when the user
        // creates it again.
        selectTool.object.addListener('dragend', function (event) {
            selectTool.mapCenter = self.gMap.getCenter();
        });
        
        if (centerTo != null) {
            selectTool.mapCenter = centerTo;
            // self.gMap.setZoom(Math.max(14,self.gMap.getZoom()));
            self.centerMap(centerTo.lng(),centerTo.lat());
        }
    }


    /////////////////////////////////////////////////////////////////////////////
    // 
    this.onSelectToolApply = function(apply)
    {
        if (apply) {

            // Get the coordinates of the shape and create the
            // call for IMWS

            var params = {
            }

            // If we work with the file window, set the scope to all files in the fw.
            if (appCore.appConfig.selectScope == 'scope') {
                params.idlist = IMatch.idlist.fileWindowFiles;
            }

            // the options parameter looks different for each shape type
            var options = {
                type: selectTool.objectType,
            }

            if (selectTool.objectType == 'circle') {
                // Needs a center coordinate and a radius
                options.center = {
                    lat: selectTool.object.getCenter().lat(),
                    lng: selectTool.object.getCenter().lng(),
                }
                options.radius = selectTool.object.getRadius();
            }
            else if (selectTool.objectType == 'rectangle') {
                // Needs the northEast and southWest parameters.
                var bounds = selectTool.object.getBounds();
                options.northEast = {
                    lat: bounds.getNorthEast().lat(),
                    lng: bounds.getNorthEast().lng()
                }
                options.southWest = {
                    lat: bounds.getSouthWest().lat(),
                    lng: bounds.getSouthWest().lng()
                }
            }
            else if (selectTool.objectType == 'polygon') {
                // We need all points in the options.points array
                options.points = [];

                selectTool.object.getPath().forEach(function(n) {
                    options.points.push({
                        lat: n.lat(), 
                        lng: n.lng()
                    });
                });
            }

            if (appCore.appConfig.selectToolUseShownCoordinates) {
                options.useShown = true;
            }

            params.options = JSON.stringify(options);

            IMWS.get('v1/search/geoarea',params).then(function(response) {

                if (response.result == 'ok') {

                    if (response.id.length == 0) {
                        // If we did find nothing, tell the user.
                        var m = new IMatchModals;
                        m.showMessageBoxOK({
                            headline: IMatchTranslate.getStr('MSGBOX_SELECT_NO_RESULT_TITLE'),
                            content: IMatchTranslate.getStr('MSGBOX_SELECT_NO_RESULT_TEST')
                        });
                    }
                    else {
                        if (appCore.appConfig.selectScope == 'scope') {
                            // If we have searched the scope, we select all files found.
                            IMatch.selectFiles({
                                id: response.id.join(',')
                            });
                        }
                        else {
                            // Else we open the results in a result window.
                            if (response.id.length > 0) {
                                var files = response.id.join(',');
                                IMatch.openResultWindow({
                                    name: IMatchTranslate.getStr('TB_SELECT_RESULT_WINDOW'),
                                    id: files
                                }).then(function(response) {
                                    // After the result window is open, select all files.
                                    IMatch.selectFiles({
                                        id: files
                                    });
                                });
                            }
                        }
                    }
                }
            });
        }

        removeSelectTool();
    }


    /////////////////////////////////////////////////////////////////////////////
    // 
    function removeSelectTool()
    {
        if (selectTool.object != null) {

            // We remember the last coordinate set for the shape for this session.
            if (selectTool.objectType == 'circle') {
                selectTool.circle = {}
                selectTool.circle.center = selectTool.object.getCenter();
                selectTool.circle.radius = selectTool.object.getRadius();
            }
            else if (selectTool.objectType == 'rectangle') {
                selectTool.rect = {}
                selectTool.rect.bounds = selectTool.object.getBounds();
            }
            else if (selectTool.objectType == 'polygon') {
                selectTool.poly = {}
                selectTool.poly.path = selectTool.object.getPath();
            }

            selectTool.object.setMap(null);
            selectTool.object = null;
            selectTool.objectType = '';
        }

        appCore.switchToolbar('default');        
    }


    /////////////////////////////////////////////////////////////////////////////
    // Center to the target marker
    this.onLocateTargetMarker = function() {
        if (markerTarget) {
            self.gMap.panTo(markerTarget.getPosition());
        }
    }


    /////////////////////////////////////////////////////////////////////////////
    // Goto the location with the given GUID
    this.onGotoLocation = function(guid) {
        for (var i = 0; i < locations.length; ++i) {
            if (locations[i].imatchLocation.guid == guid) {
                var center = new google.maps.LatLng(locations[i].imatchLocation.lat, locations[i].imatchLocation.lng);
                self.gMap.panTo(center);
                break;
            }
        }
    }


    /////////////////////////////////////////////////////////////////////////////
    // appCore has determined that a file was changed. 
    // We check if the file is currently on the map and update it.
    this.onFileUpdate = function(msg) {
        var hasfile = false;
        for (var i = 0; i < files.length; ++i) {
            if (files[i].id == msg.d1) {
                hasfile = true;
                break;
            }
        }

        if (hasfile) {
            loadFilesDelayed();
        }
    }

    
    /////////////////////////////////////////////////////////////////////////////
    // Locations in the database were changed (by this app or something else).
    this.onLocationUpdate = function(msg)
    {
        loadLocations();
    }


    /////////////////////////////////////////////////////////////////////////////
    // Called when the state of the lock option changes
    this.onLockChanged = function()
    {
        locations.forEach(function(l) {
            l.setEditable(!appCore.appConfig.lockMove);
            l.setDraggable(!appCore.appConfig.lockMove);
        });

        self.markers.forEach(function(m) {
            m.setDraggable(!appCore.appConfig.lockMove);
        });
    }


   /////////////////////////////////////////////////////////////////////////////
   // A track file was loaded.
   this.onTrackLoaded = function(fitTrack)
   {
        clearTrackLogs();

        if (!appCore.tracks().hasTracks()) return;

        var trackData = appCore.tracks().trackData();

        // Thickness of the polyline
        var lineWeight = 6;

        // The marker icon
        var icon = {
            url: 'img/waypoint.png',
            anchor: new google.maps.Point(0,(lineWeight+4)/2),
            scaledSize: new google.maps.Size(lineWeight+4,lineWeight+4),
        }

        // var showMarker = self.gMap.getZoom() >= trackLogMarkerMinZoom;

        trackBounds = new google.maps.LatLngBounds();

        for (var i = 0; i < trackData.tracks.length; ++i) {

            // The points for the polygon
            var path = [];
            var markers = [];
            var title;

            // The bounds of the polygon. We calculate this only once.
            for (var p = 0; p < trackData.tracks[i].points.length; ++p) {
                path.push({
                    lat: trackData.tracks[i].points[p].lat,
                    lng: trackData.tracks[i].points[p].lon
                });
                trackBounds.extend(path[path.length-1]);

                title = trackData.tracks[i].points[p].name;
                if (title != '') {
                    title += '\n';
                }
                title += trackData.tracks[i].points[p].time;

                /*
                markers.push(new google.maps.Marker({
                    visible: showMarker,
                    map: self.gMap,
                    icon: icon,
                    title: title,
                    position: new google.maps.LatLng(trackData.tracks[i].points[p].lat, trackData.tracks[i].points[p].lon)
                }));
                */
            }

            var p = new google.maps.Polyline({
                path: path,
                geodesic: true,
                strokeColor: '#0000ff',
                strokeOpacity: 0.5,
                strokeWeight: lineWeight,
                map: self.gMap
            });
            
            trackLogs.push({
                name: trackData.tracks[i].name,
                poly: p,
                // markers: markers,
            });
        }

        // Make sure the track is visible. Zoom the map to contain the entire path.
        if (trackLogs.length > 0) {
            if (fitTrack) {
                self.gMap.fitBounds(trackBounds);
            }
        }

   }


   /////////////////////////////////////////////////////////////////////////////
   // Center to the bounding box of the track.
   this.onCenterTrack = function()
   {
        if (trackLogs.length > 0) {
           self.gMap.fitBounds(trackBounds);
        }
   }

   /////////////////////////////////////////////////////////////////////////////
   // Show files contained in the current track. These files have no GPS data
   // set, the GPS data only exists in the file objects maintained by the track.
   this.onShowTrackFiles = function() {
        clearTrackMarkers();

        appCore.tracks().trackData().files.forEach(function(f) {

            if (f.mapped) {
                var options = {
                    map: self.gMap,
                    file: f,
                    label: f.name,
                    title: f.mappedGPS.lat.toFixed(6) + ', ' + f.mappedGPS.lng.toFixed(6) + '\n' + (f.mappedGPS.alt).toFixed(2) + 'm\n' + moment(f.mappedGPS.time).format(),
                    position: new google.maps.LatLng(f.mappedGPS.lat, f.mappedGPS.lng),
                    draggable: false,
                    icon : {
                        url: 'img/marker3.png',
                        labelOrigin: new google.maps.Point(0,40)
                    }         
                }

                var marker = new google.maps.Marker(options);
                trackMarkers.push(marker);
            }

        }, this);
   }


   /////////////////////////////////////////////////////////////////////////////
   // Remove track markers
   this.onHideTrackFiles = function()
   {
       clearTrackMarkers();
   }


   /////////////////////////////////////////////////////////////////////////////
   // Remove all track markers from the map
   function clearTrackMarkers()
   {
        for (var i = 0; i < trackMarkers.length; ++i) {
            trackMarkers[i].setMap(null);
        }
        trackMarkers = [];
   }


   /////////////////////////////////////////////////////////////////////////////
   // A track file was unloaded
   this.onTrackUnLoaded = function()
   {
        clearTrackLogs();
   }


   /////////////////////////////////////////////////////////////////////////////
   // Clear all track logs
   function clearTrackLogs()
   {
       clearTrackMarkers();       
       for (var i = 0; i < trackLogs.length; ++i) {
           trackLogs[i].poly.setMap(null);
        //    for (var m = 0; m < trackLogs[i].markers.length; m++) {
        //        trackLogs[i].markers[m].setMap(null);
        //    }
       }
       
       trackLogs = [];
       trackBounds = null;
   }


   /////////////////////////////////////////////////////////////////////////////
   // Close open popups
   this.onClosePopups = function()
   {
       closePopup();
   }


   /////////////////////////////////////////////////////////////////////////////
   // This is called by the appCore when it destroys the map
   this.onDestroy = function()
   {
        saveConfig();
   }

    /////////////////////////////////////////////////////////////////////////////
    // Update user settings and save back
    function saveConfig()
    {
        if (self.gMap) {
            appCore.appConfig.zoomLevel = self.gMap.getZoom();
            var p = self.gMap.getCenter();
            appCore.appConfig.mapCenter.lng = p.lng();
            appCore.appConfig.mapCenter.lat = p.lat();
            appCore.saveConfig();
        }
    }

    
    /////////////////////////////////////////////////////////////////////////////
    // Called by appCore when the user clicks on the appy target marker command.
    this.onApplyTargetMarker = function()
    {
        if (markerTarget) {

            function apply()
            {
                clearUndo();
                IMatch.setGPSCoordinates({
                    target: 'created',
                    lat: markerTarget.getPosition().lat(),
                    lng: markerTarget.getPosition().lng(),
                    idlist: IMatch.idlist.fileWindowSelection
                }).then(function(response) {
                    // If the update was successful, we reload the files.
                    self.loadFiles(true);
                },error => {
                    appCore.showError('setGps',error);
                });
            }

            if (appCore.sessionSettings.showApplyTargetMarkerPromt) {
                // Ask the user
                var m = new IMatchModals();
                m.showModal(`modals/modal-apply-target-marker-tpl.html`,undefined,function() {
                    // OK
                    apply();
                },
                // Cancel
                undefined,
                // Bind handler after the modal has been created.
                function () {
                    $('#apply-target-session-warning').click(function() {
                        appCore.sessionSettings.showApplyTargetMarkerPromt = false;
                    });
                });
            }
            else {
                apply();
            }
        }
    }


    /////////////////////////////////////////////////////////////////////////////
    // Apply the coordinates and data of a location
    this.onApplyLocation = function(guid)
    {
        // Ask the user
        var m = new IMatchModals();
        m.showMessageBoxYesNo({
            headline: IMatchTranslate.getStr('MSGBOX_APPLY_LOC_TITLE'),
            content: IMatchTranslate.getStr('MSGBOX_APPLY_LOC_TEXT'),
            okCallback: function() {
                clearUndo();
                IMWS.post('v1/geolocation/apply',{
                    guid: guid,
                    idlist: IMatch.idlist.fileWindowSelection
                }).then(function(response) {
                    self.loadFiles(true);
                });
            }
        });
    }


    /////////////////////////////////////////////////////////////////////////////
    //
    this.onFindNearbyLocation = function(guid) {
        if (locations) {
            locations.forEach(function(l) {
                if (l.imatchLocation.guid == guid) {
                    IMatch.findNearbyFiles({
                        lat: l.imatchLocation.lat,
                        lng: l.imatchLocation.lng,
                        alt: 0                
                    }).then(function(response) {
                        if (response.result == 'ok') {
                            appCore.openFindNearbyResultWindow(IMatchTranslate.getStr('MAP_FIND_RESULTWINDOW'),response.id);
                        }
                    });
                }
            });
        }
    }


    /////////////////////////////////////////////////////////////////////////////
    //
    function setTargetMarker(lng,lat)
    {
        if (markerTarget == null) {
            markerTarget = new google.maps.Marker({
                map: self.gMap,
                icon: {
                    url: 'img/target2.svg',
                    scaledSize: new google.maps.Size(24,40),
                },
                animation: google.maps.Animation.DROP,
                position: new google.maps.LatLng(lat, lng),
                draggable:true
            });

            markerTarget.addListener('rightclick', function (event) {
                closePopup();
                createTargetPopup(event.latLng);
            });

             markerTarget.addListener('dragend', function (event) {
                 //console.log(event);
            });

            appCore.onTargetMarkerChanged(true);
        }
        else {
            markerTarget.setPosition(new google.maps.LatLng(lat, lng));
        }
    }

    //////////////////////////////////////////////////////
    // This method produces the data for the popup element
    function createTargetPopup(latLng, feature) {

        closePopup();

        var content = [
            '<div id="target-popup-gmaps" class="ptc-map-popup-gmaps container-fluid">',
            '  <h4 imt-translate>Target Marker</h4>',
            '  <div class="row">',
            '    <div class="col-sm-12 ptc-popup-row"><i class="fa fa-check-square fa-lg btn-color-target imatch-margin-r-10"></i><a href="#" id="btn-target-apply" data-imt-translate="MAP_POPUP_TARGET_APPLY"></a></div>',
            '  </div>',
            '  <div class="row">',
            '    <div class="col-sm-12 ptc-popup-row"><img src="img/location.svg" class="imatch-margin-r-10" style="height:1.25em; width:auto;"><a href="#" id="btn-target-create-loc" data-imt-translate="MAP_POPUP_TARGET_CREATE_LOC"></a></div>',
            '  </div>',
            '  <div class="row">',
            '    <div class="col-sm-12 ptc-popup-row"><i class="fa fa-search fa-lg imatch-margin-r-10"></i><a href="#" id="btn-target-find-nearby" data-imt-translate="MAP_POPUP_BTN_FINDNEARBY"></a></div>',
            '  </div>',
            '  <div class="row">',
            '    <div class="col-sm-12 ptc-popup-row"><img src="img/location-remove.svg" class="imatch-margin-r-10" style="height:1.25em; width:auto;"><a href="#" id="btn-target-remove-marker" data-imt-translate="MAP_POPUP_TARGET_REMOVE_MARKER"></a></div>',
            '  </div>',
            '</div>'
        ].join('\n');

        mapPopup.infoWnd.setContent(content);
        mapPopup.infoWnd.setOptions({});
        mapPopup.infoWnd.open(self.gMap, markerTarget);

        // Translate the contents in the popup after it has been added to the DOM.
        IMatchTranslate.translateElement('#target-popup-gmaps');

        // Button handlers

        $('#btn-popup-close').click(function(e) {
            closePopup();
            e.preventDefault();
            return false;
        })

        $('#btn-target-apply').click(function(e) {
            closePopup();
            self.onApplyTargetMarker();
            return false;
        })
        
        $('#btn-target-create-loc').click(function(e) {
            closePopup();
            IMatch.createGeoLocation({
                lat: markerTarget.getPosition().lat(),
                lng: markerTarget.getPosition().lng(),
                templatefile: 0
            });
            return false;
        })

        $('#btn-target-remove-marker').click(function(e) {
            closePopup();
            if (markerTarget) {
                markerTarget.setMap(null);
                markerTarget = null;
                appCore.onTargetMarkerChanged(false);
            }
            return false;
        })

        $('#btn-target-find-nearby').click(function(e) {
            closePopup();
            IMatch.findNearbyFiles({
                lat: markerTarget.getPosition().lat(),
                lng: markerTarget.getPosition().lng(),
                alt: 0                
            }).then(function(response) {
                appCore.openFindNearbyResultWindow(IMatchTranslate.getStr('MAP_FIND_RESULTWINDOW'),response.id);
            });
            return false;
        })

        return true;
    }    


    //////////////////////////////////////////////////////
    // This method produces the data for the popup element
    function createLocationPopup(latLng, shape) {

        closePopup();

        var content = [
            '<div id="target-popup-gmaps" class="ptc-map-popup-gmaps container-fluid">',
            '  <h4>{{loc}}</h4>',
            '  <div class="row">',
            '    <div class="col-sm-12 ptc-popup-row"><i class="fa fa-pencil fa-lg imatch-margin-r-10"></i><a href="#" id="btn-loc-edit" data-imt-translate="MAP_POPUP_LOC_EDIT"></a></div>',
            '  </div>',
            '  <div class="row">',
            '    <div class="col-sm-12 ptc-popup-row"><i class="fa fa-check-square fa-lg btn-color-target imatch-margin-r-10"></i><a href="#" id="btn-loc-apply" data-imt-translate="MAP_POPUP_LOC_APPLY"></a></div>',
            '  </div>',
            '  <div class="row">',
            '    <div class="col-sm-12 ptc-popup-row"><i class="fa fa-search fa-lg imatch-margin-r-10"></i><a href="#" id="btn-loc-find-nearby" data-imt-translate="MAP_POPUP_BTN_FINDNEARBY"></a></div>',
            '  </div>',
            '  <div class="row">',
            '    <br/><div class="col-sm-12 ptc-popup-row"><i class="fa fa-times-circle fa-lg btn-color-delete imatch-margin-r-10"></i><a href="#" id="btn-loc-delete" data-imt-translate="MAP_POPUP_LOC_DELETE"></a></div>',
            '  </div>',
            '  <div class="row">',
            '    <div class="btn-toolbar pull-right">',
            '      <button id="btn-popup-close" class="btn btn-default btn-sm" style="margin-left:20px;" data-imt-title="MAP_POPUP_BTN_CLOSE"><i class="fa fa-times"></i></button>',
            '    </div>',
            '  </div>',
            '</div>'
        ].join('\n');

        content = content.replace('{{loc}}',shape.imatchLocation.name);

        mapPopup.infoWnd.setContent(content);
        mapPopup.infoWnd.setOptions({});
        mapPopup.infoWnd.open(self.gMap);
        mapPopup.infoWnd.setPosition(latLng);
        mapPopup.locationShape = shape;

        // Translate the contents in the popup after it has been added to the DOM.
        IMatchTranslate.translateElement('#target-popup-gmaps');

        // Button handlers

        $('#btn-popup-close').click(function(e) {
            closePopup();
            e.preventDefault();
            return false;
        })

        $('#btn-loc-apply').click(function(e) {
            closePopup();
            self.onApplyLocation(mapPopup.locationShape.imatchLocation.guid);
            return false;
        })

        $('#btn-loc-find-nearby').click(function(e) {
            closePopup();
            self.onFindNearbyLocation(mapPopup.locationShape.imatchLocation.guid);
            return false;
        });
        
        
        $('#btn-loc-edit').click(function(e) {
            closePopup();

            IMatch.editGeoLocation({
                guid: mapPopup.locationShape.imatchLocation.guid
            });

            e.preventDefault();
            return false;
        });

        $('#btn-loc-delete').click(function(e) {
            closePopup();
            // Ask the user
            var m = new IMatchModals();
            m.showMessageBoxYesNo({
                headline: IMatchTranslate.getStr('MSGBOX_DELETE_LOC_TITLE'),
                content: IMatchTranslate.getStr('MSGBOX_DELETE_LOC_TEXT'),
                okCallback: function() {
                    clearUndo();
                    IMWS.delete('v1/geolocation',{
                        guid: mapPopup.locationShape.imatchLocation.guid
                    });
                }
            });

            e.preventDefault();
            return false;
        });


        return true;
    }    


    /////////////////////////////////////////////////////////////////////////////
    // Creates a destination coordinate pair for the file in the given marker.
    function createDirection(marker) {
        var pos = mapPopup.markers[mapPopup.index].getPosition();

        // We place the marker at 1/3 of the displayed map area
        var dist = (self.gMap.getBounds().getNorthEast().lng() - self.gMap.getBounds().getSouthWest().lng()) * 0.33;

        MapUtility.createDest(pos.lng(),pos.lat(),marker.file, dist).then(response => {
            // Fetch the data of the updated file. 
            // IMatch will fill in image direction and other values.
            IMWS.get('v1/files',{
                id: marker.file.id,
                fields: 'gps'
            }).then(response => {
                marker.file.gps = response.files[0].gps;
                createDestinationMarker(marker);
                createFOV();
            });
        });
    }

    
    /////////////////////////////////////////////////////////////////////////////
    // Position the target marker on the coordinates specified in text form.
    this.positionTargetMarker = function (lng,lat)
    {
        setTargetMarker(lng,lat);
        self.centerMap(lng,lat);
    }


    /////////////////////////////////////////////////////////////////////////////
    // Center the map at the given coordinates
    this.centerMap = function(lng,lat)
    {
        var center = new google.maps.LatLng(lat, lng)
        self.gMap.panTo(center);        
    }    


}

    